"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.check = exports.checkPrimitive = exports.preserveFeatureSite = exports.checkByRegex = exports.makeContext = exports.makeCache = void 0;
var lineColumn = require("line-column");
var LruCache = require("lru-cache");
var parser = require("@solidity-parser/parser");
function makeCache(size) {
    return new LruCache(size);
}
exports.makeCache = makeCache;
function makeContext(codeInfo, cacheSize) {
    if (cacheSize === void 0) { cacheSize = 10; }
    return {
        codeInfo: codeInfo,
        cache: makeCache(cacheSize)
    };
}
exports.makeContext = makeContext;
function checkByRegex(context, regex) {
    var result = [];
    var execArray;
    while ((execArray = regex.exec(context.codeInfo.code)) != null) {
        result.push({ literal: execArray[0], index: execArray.index });
    }
    return result;
}
exports.checkByRegex = checkByRegex;
function getLineColumnFinderFromCache(context) {
    if (!context.cache.has(context.codeInfo.filename)) {
        context.cache.set(context.codeInfo.filename, lineColumn(context.codeInfo.code));
    }
    return context.cache.get(context.codeInfo.filename);
}
function preserveFeatureSite(context, node) {
    if (node.loc === undefined) { // sometimes location is not given by @solidity-parser/parser
        return {};
    }
    var lineColumnFinder = getLineColumnFinderFromCache(context);
    var location = node.loc; // node.loc is wrongly typed in @solidity-parser/parser
    var startIndex = lineColumnFinder.toIndex({ line: location.start.line, column: location.start.column + 1 });
    var endIndex = lineColumnFinder.toIndex({ line: location.end.line, column: location.end.column + 1 });
    return {
        literal: context.codeInfo.code.substring(startIndex, endIndex + 1),
        location: location,
        index: startIndex,
    };
}
exports.preserveFeatureSite = preserveFeatureSite;
function checkPrimitive(context, primitiveVisitor) {
    parser.visit(context.codeInfo.ast, primitiveVisitor);
}
exports.checkPrimitive = checkPrimitive;
function check(context, structuredVisitorOrNodeName) {
    if (typeof structuredVisitorOrNodeName === 'string') { // structuredVisitorOrNodeName is ExtendedASTNodeTypeString
        return checkByASTNodeName(context, structuredVisitorOrNodeName);
    }
    else if (Array.isArray(structuredVisitorOrNodeName)) { // structuredVisitorOrNodeName is ExtendedASTNodeTypeString[]
        return checkByASTNodeNames(context, structuredVisitorOrNodeName);
    }
    else { // structuredVisitorOrNodeName is StructuredVisitor
        return checkByStructuredVisitor(context, structuredVisitorOrNodeName);
    }
}
exports.check = check;
function checkByASTNodeName(context, nodeTypeName) {
    return new Promise(function (resolve) {
        var _a;
        var result = [];
        parser.visit(context.codeInfo.ast, (_a = {},
            _a[nodeTypeName] = function (node) {
                result.push(preserveFeatureSite(context, node));
            },
            _a['SourceUnit:exit'] = function () {
                resolve(result);
            },
            _a));
    });
}
function checkByASTNodeNames(context, nodeTypeNames) {
    return new Promise(function (resolve) {
        var result = [];
        var visitor = {
            'SourceUnit:exit': function () {
                resolve(result);
            },
        };
        for (var _i = 0, nodeTypeNames_1 = nodeTypeNames; _i < nodeTypeNames_1.length; _i++) {
            var names = nodeTypeNames_1[_i];
            visitor[names] = (function (node) { return result.push(preserveFeatureSite(context, node)); });
        }
        parser.visit(context.codeInfo.ast, visitor);
    });
}
function checkByStructuredVisitor(context, structuredVisitor) {
    return new Promise(function (resolve) {
        var primitiveVisitor = buildPrimitiveVisitor(structuredVisitor, function (node) { return preserveFeatureSite(context, node); }, function (featureSites) { return resolve(featureSites); });
        parser.visit(context.codeInfo.ast, primitiveVisitor);
    });
}
function buildPrimitiveVisitor(structuredVisitor, captureSite, onComplete) {
    var runTimeContextNodeRecord = {};
    var runTimeContextNodeTypePaths = [];
    var buildTimeNodeTypeStack = [];
    var capturedFeatureSites = [];
    var primitiveVisitor = {
        'SourceUnit:exit': function () {
            onComplete(capturedFeatureSites);
        },
    };
    function buildPrimitiveVisitor(structuredVisitor, root) {
        var _loop_1 = function (nodeType, handler) {
            var probeObject = handler();
            if (typeof probeObject === 'boolean' || typeof probeObject === 'undefined') { // handler is StructuredVisitorFinalHandler
                runTimeContextNodeTypePaths.push(buildTimeNodeTypeStack.slice());
                if (primitiveVisitor[nodeType] === undefined) {
                    primitiveVisitor[nodeType] = function (finalNode) {
                        var _a;
                        for (var _i = 0, runTimeContextNodeTypePaths_1 = runTimeContextNodeTypePaths; _i < runTimeContextNodeTypePaths_1.length; _i++) {
                            var contextNodeTypePath = runTimeContextNodeTypePaths_1[_i];
                            var steppingVisitor = root;
                            for (var _b = 0, contextNodeTypePath_1 = contextNodeTypePath; _b < contextNodeTypePath_1.length; _b++) {
                                var contextNodeType = contextNodeTypePath_1[_b];
                                var contextNode = runTimeContextNodeRecord[contextNodeType];
                                if (contextNode === null) {
                                    steppingVisitor = null;
                                    break;
                                }
                                else {
                                    steppingVisitor = steppingVisitor[contextNodeType](contextNode);
                                }
                            }
                            if (steppingVisitor !== null) {
                                var finalMatch = (_a = steppingVisitor[nodeType](finalNode)) !== null && _a !== void 0 ? _a : false;
                                if (finalMatch) {
                                    capturedFeatureSites.push(captureSite(finalNode));
                                    return;
                                }
                            }
                        }
                    };
                }
            }
            else { // handler is StructuredVisitorIntermediateHandler
                if (primitiveVisitor[nodeType] === undefined) {
                    primitiveVisitor[nodeType] = function (node) {
                        runTimeContextNodeRecord[nodeType] = node;
                    };
                }
                var exitKey = nodeType + ":exit";
                if (primitiveVisitor[exitKey] === undefined) {
                    primitiveVisitor[exitKey] = function () {
                        runTimeContextNodeRecord[nodeType] = null;
                    };
                }
                buildTimeNodeTypeStack.push(nodeType);
                buildPrimitiveVisitor(probeObject, root);
                buildTimeNodeTypeStack.pop();
            }
        };
        for (var _i = 0, _a = Object.entries(structuredVisitor); _i < _a.length; _i++) {
            var _b = _a[_i], nodeType = _b[0], handler = _b[1];
            _loop_1(nodeType, handler);
        }
    }
    buildPrimitiveVisitor(structuredVisitor, structuredVisitor);
    return primitiveVisitor;
}
